/*====================================================================
 * linux/drivers/ide/pci/initio.c		Version 1.00	Dec 2, 2003
 * Modify Files:
 *	1. linux/drivers/ide/Config.in -
 *		add CONFIG_BLK_DEV_INITIO under CONFIG_BLK_DEV_IDEDMA_PCI(line 74)
 *	2. linux/Documentation/Configure.help -
 *		add line 1292
 *	3. linux/drivers/ide/pci/Makefile -
 *		add initio.o(line 19)
 *	4. linux/include/linux/pci_ids.h -
 *		add line 963
 *====================================================================*/

#include <linux/config.h>
#include <linux/types.h>
#include <linux/module.h>
#include <linux/pci.h>
#include <linux/delay.h>
#include <linux/hdreg.h>
#include <linux/ide.h>
#include <linux/init.h>

#include <asm/io.h>

#include "ide_modes.h"
#include "initio.h"

#if defined(DISPLAY_INITIO_TIMINGS) && defined(CONFIG_PROC_FS)
#include <linux/stat.h>
#include <linux/proc_fs.h>

static u8 initio_proc = 0;
#define INITIO_MAX_DEVS		16
static struct pci_dev *initio_devs[INITIO_MAX_DEVS];
static int n_initio_devs;

/*==================================================================
 *	initio_selreg -	return register base
 *	@hwif: interface
 *	@r: config offset
 *
 *	Turn a config register offset into the right address in either
 *	PCI space or MMIO space to access the control register in question
 *	Thankfully this is a configuration operation so isnt performance
 *	criticial. 
 *===================================================================*/
static unsigned long initio_selreg(ide_hwif_t *hwif, int r)
{
	unsigned long	base;

	base = (unsigned long)hwif->hwif_data + (unsigned long)r;
	if (hwif->channel)
		base += 0x40;

	return base;
}
	
	
/*==================================================================
 *	print_initio_get_info	-	print minimal proc information
 *	@buf: buffer to write into (kernel space)
 *	@dev: PCI device we are describing
 *	@index: Controller number
 *
 *	Print the basic information for the state of the CMD680/SI3112
 *	channel. We don't actually dump a lot of information out for
 *	this controller although we could expand it if we needed.
 ===================================================================*/
 
static char *print_initio_get_info (char *buf, struct pci_dev *dev, int index)
{
	char *p		= buf;
	u8 mmio		= (pci_get_drvdata(dev) != NULL) ? 1 : 0;
	unsigned long bmdma = pci_resource_start(dev, 4);
	
	if(mmio)
		bmdma = pci_resource_start(dev, 5);

	p += sprintf(p, "\nController: %d\n", index);
	p += sprintf(p, "Inic%x Chipset.\n", dev->device);
	if (mmio)
		p += sprintf(p, "MMIO Base 0x%lx\n", bmdma);
	p += sprintf(p, "%s-DMA Base 0x%lx\n", (mmio)?"MMIO":"BM", bmdma);
	p += sprintf(p, "%s-DMA Base 0x%lx\n", (mmio)?"MMIO":"BM", bmdma+0x40);
	return (char *)p;
}

/*==================================================================
 *	initio_get_info	-	proc callback
 *	@buffer: kernel buffer to complete
 *	@addr: written with base of data to return
 *	offset: seek offset
 *	count: bytes to fill in 
 *
 *	Called when the user reads data from the virtual file for this
 *	controller from /proc
 ===================================================================*/
 
static int initio_get_info (char *buffer, char **addr, off_t offset, int count)
{
	char *p = buffer;
	int len;
	u16 i;

	p += sprintf(p, "\n");
	for (i = 0; i < n_initio_devs; i++) {
		struct pci_dev *dev	= initio_devs[i];
		p = print_initio_get_info(p, dev, i);
	}
	/* p - buffer must be less than 4k! */
	len = (p - buffer) - offset;
	*addr = buffer + offset;
	
	return len > count ? count : len;
}

#endif	/* defined(DISPLAY_INITIO_TIMINGS) && defined(CONFIG_PROC_FS) */

/**********************
 * Support Functions  *
 **********************/
/*=======================================================================
 *	init_mmio_iops_initio - set up the iops for MMIO
 *	@hwif: interface to set up
 *
 *	The basic setup here is fairly simple, we can use standard MMIO
 *	operations. However we do have to set the taskfile register offsets
 *	by hand as there isnt a standard defined layout for them this
 *	time.
 *
 *	The hardware supports buffered taskfiles and also some rather nice
 *	extended PRD tables. Unfortunately right now we don't.
 *=======================================================================*/
static void __init init_mmio_iops_initio(ide_hwif_t *hwif)
{
	struct pci_dev	*dev=hwif->pci_dev;
	unsigned long	base;
	u8				ch=hwif->channel;
	hw_regs_t		hw;

	default_hwif_mmiops(hwif);
	memset(&hw, 0, sizeof(hw_regs_t));

	base = (unsigned long)pci_get_drvdata(dev);
	hw.priv	= (void *)base;
	hw.irq = hwif->pci_dev->irq;

	printk("           - R1x7 = 0x%02X\n", hwif->INB(base+7));

	if (ch == 1)
		base += 0x40;
	hw.io_ports[IDE_DATA_OFFSET]	= base;
	hw.io_ports[IDE_ERROR_OFFSET]	= base + 0x01;
	hw.io_ports[IDE_NSECTOR_OFFSET]	= base + 0x02;
	hw.io_ports[IDE_SECTOR_OFFSET]	= base + 0x03;
	hw.io_ports[IDE_LCYL_OFFSET]	= base + 0x04;
	hw.io_ports[IDE_HCYL_OFFSET]	= base + 0x05;
	hw.io_ports[IDE_SELECT_OFFSET]	= base + 0x06;
	hw.io_ports[IDE_STATUS_OFFSET]	= base + 0x07;
//	hw.io_ports[IDE_CONTROL_OFFSET]	= base + 0x08;
	hw.io_ports[IDE_IRQ_OFFSET]	    = base + 0x16;

	hw.sata_scr[SATA_STATUS_OFFSET]	= base + 0x20;
	hw.sata_scr[SATA_ERROR_OFFSET]	= base + 0x24;
	hw.sata_scr[SATA_CONTROL_OFFSET]= base + 0x28;
//	hw.sata_misc[SATA_MISC_OFFSET]	= 0;
	hw.sata_misc[SATA_PHY_OFFSET]	= base + 0x38;	//Base 4/5 Page 1
	hw.sata_misc[SATA_IEN_OFFSET]	= base + 0x0A;

	memcpy(&hwif->hw, &hw, sizeof(hw));
	memcpy(hwif->io_ports, hwif->hw.io_ports, sizeof(hwif->hw.io_ports));
	memcpy(hwif->sata_scr, hwif->hw.sata_scr, sizeof(hwif->hw.sata_scr));
	memcpy(hwif->sata_misc, hwif->hw.sata_misc, sizeof(hwif->hw.sata_misc));

	hwif->irq = hw.irq;
	hwif->hwif_data	= (void *)hw.priv;
	hwif->dma_base = base + 0x0B;
	hwif->dma_base2	= 0;
	hwif->mmio = 1;
	hwif->addressing = 1;
}


static unsigned int setup_mmio_initio(struct pci_dev *dev, const char *name)
{
	u32		bar5 = pci_resource_start(dev, 5);
	u32		barsize = pci_resource_len(dev, 5);
	u32		addr;
//	u16		val;
//	int		i=10;
	void	*ioaddr;

	/*
	 *	Drop back to PIO if we can't map the mmio. Some
	 *	systems seem to get terminally confused in the PCI
	 *	spaces.
	 */
	 
	if(!request_mem_region(bar5, barsize, name))
	{
		printk(KERN_WARNING "inic1622: IDE controller MMIO ports not available.\n");
		return 0;
	}
		
	ioaddr = ioremap(bar5, barsize);

	printk("   ioremap 0x%08lX to 0x%08lX, length=0x%08lX\n", 
			(unsigned long)bar5, (unsigned long)ioaddr, (unsigned long)barsize);

	if (ioaddr == NULL)
	{
		release_mem_region(bar5, barsize);
		return 0;
	}

	pci_set_master(dev);
	pci_set_drvdata(dev, ioaddr);
	addr = (unsigned long)ioaddr + 0x7C;

//==========================================================================
/*	printk("   0. Before RESET CHIP: R1x7 = 0x%02X\n", readb(ioaddr+7));

	// Disable Global Interrupt & Reset entire chip
	val = readw(addr);
	printk("   1. Disable Global Interrupt & Reset chip\n");
	writew((val|0x2100), addr);
	do {
		udelay(10);
		i--;
	}while((readw(addr)&SOFTRST) && (i!=0));
	if (i == 0)
	{
		printk("   ### Chip software reset fail!\n");
		return 0;
	}
	
	// Enable Global interrupt & Set FIFO threshold
	printk("   2. Enable Global Interrupt & Set FIFO threshold\n");
	writew(0x0800, addr);

	printk("   3. After RESET CHIP: R1x7 = 0x%02X\n", readb(ioaddr+7));*/
//==========================================================================

	// Enable Global interrupt & Set FIFO threshold
	printk("   - Enable Global Interrupt & Set FIFO threshold\n");
	writew(0x0800, addr);

#if defined(DISPLAY_INITIO_TIMINGS) && defined(CONFIG_PROC_FS)
	initio_devs[n_initio_devs++] = dev;

	if (!initio_proc)
	{
		initio_proc = 1;
		ide_pci_register_host_proc(&initio_procs[0]);
	}
#endif /* DISPLAY_INITIO_TIMINGS && CONFIG_PROC_FS */

	return 1;
}


/*============================================================
 *	config_chipset_for_dma	-	configure for DMA
 *	@drive: drive to configure
 *
 *	Called by the IDE layer when it wants the timings set up.
 *============================================================*/
static int config_chipset_for_dma(ide_drive_t *drive)
{
	ide_hwif_t	*hwif=HWIF(drive);
	u8			speed=0;

	if ((drive->id->dma_ultra & 0x0040) && (drive->id->dma_ultra & hwif->ultra_mask))
		speed = XFER_UDMA_6;
	else if ((drive->id->dma_ultra & 0x0020) && (drive->id->dma_ultra & hwif->ultra_mask))
		speed = XFER_UDMA_5;
	else if ((drive->id->dma_ultra & 0x0010) && (drive->id->dma_ultra & hwif->ultra_mask))
		speed = XFER_UDMA_4;
	else if ((drive->id->dma_ultra & 0x0008) && (drive->id->dma_ultra & hwif->ultra_mask))
		speed = XFER_UDMA_3;
	else if ((drive->id->dma_ultra & 0x0004) && (drive->id->dma_ultra & hwif->ultra_mask))
		speed = XFER_UDMA_2;
	else if ((drive->id->dma_ultra & 0x0002) && (drive->id->dma_ultra & hwif->ultra_mask))
		speed = XFER_UDMA_1;
	else if ((drive->id->dma_ultra & 0x0001) && (drive->id->dma_ultra & hwif->ultra_mask))
		speed = XFER_UDMA_0;
	else if ((drive->id->dma_mword & 0x0004) && (drive->id->dma_mword & hwif->mwdma_mask))
		speed = XFER_MW_DMA_2;
	else if ((drive->id->dma_mword & 0x0002) && (drive->id->dma_mword & hwif->mwdma_mask))
		speed = XFER_MW_DMA_1;
	else if ((drive->id->dma_mword & 0x0001) && (drive->id->dma_mword & hwif->mwdma_mask))
		speed = XFER_MW_DMA_0;
	else if ((drive->id->dma_1word & 0x0004) && (drive->id->dma_1word & hwif->swdma_mask))
		speed = XFER_SW_DMA_2;
	else if ((drive->id->dma_1word & 0x0002) && (drive->id->dma_1word & hwif->swdma_mask))
		speed = XFER_SW_DMA_1;
	else if ((drive->id->dma_1word & 0x0001) && (drive->id->dma_1word & hwif->swdma_mask))
		speed = XFER_SW_DMA_0;

	if (!speed)
		return 0;

	printk("      -> config_chipset_for_dma\n         -> ");

	if (ide_set_xfer_rate(drive, speed))
	{
		printk("        ==> some bug in ide_set_xfer_rate\n");
		return 0;
	}

	if (!drive->init_speed)
		drive->init_speed = speed;

	return ide_dma_enable(drive);
}


/*===============================================================
 *	initio_configure_drive_for_dma - set up for DMA transfers
 *	@drive: drive we are going to set up
 *
 *	Set up the drive for DMA, tune the controller and drive as 
 *	required. If the drive isn't suitable for DMA or we hit
 *	other problems then we will drop down to PIO and set up
 *	PIO appropriately
 *===============================================================*/
static int initio_config_drive_for_dma(ide_drive_t *drive)
{
	ide_hwif_t			*hwif=HWIF(drive);
	struct hd_driveid	*id=drive->id;

	printk("   - initio_config_drive_for_dma:\n");

	if ((id->capability & 1) == 0)
		printk("      - %s not support DMA\n", drive->name);
	if (drive->autodma == 0)
		printk("      - %s: Auto DMA off\n", drive->name);

	if ((id->capability & 1)!= 0 && drive->autodma)
	{
		/* Consult the list of known "bad" drives */
		if (hwif->ide_dma_bad_drive(drive))
		{
			printk("      - %s on BAD drive list\n", drive->name);
			goto fast_ata_pio;
		}

		if (id->field_valid & 4)
		{
			if (id->dma_ultra & hwif->ultra_mask)
			{
				/* Force if Capable UltraDMA */
				int dma = config_chipset_for_dma(drive);
				if ((id->field_valid & 2) && !dma)
				{
					printk("      - config_chipset_for_dma for UDMA fail, try MDMA\n");
					goto try_dma_modes;
				}
			}
		}
		else if (id->field_valid & 2)
		{
try_dma_modes:
			if ((id->dma_mword & hwif->mwdma_mask) ||
				(id->dma_1word & hwif->swdma_mask))
			{
				/* Force if Capable regular DMA modes */
				if (!config_chipset_for_dma(drive))
				{
					printk("      - config_chipset_for_dma for MDMA fail, set PIO 4\n");
					goto no_dma_set;
				}
			}
		}
		else if (hwif->ide_dma_good_drive(drive) &&
				 (id->eide_dma_time < 150))
		{
			/* Consult the list of known "good" drives */
			if (!config_chipset_for_dma(drive))
			{
				printk("      - config_chipset_for_dma for SDMA fail, set PIO 4\n");
				goto no_dma_set;
			}
		}
		else
		{
			printk("      - This drive not support DMA\n");
			goto fast_ata_pio;
		}
	}
	else if ((id->capability & 8) || (id->field_valid & 2))
	{
fast_ata_pio:
no_dma_set:
		ide_config_drive_speed(drive, XFER_PIO_4);
		return hwif->ide_dma_off_quietly(drive);
	}
	return hwif->ide_dma_on(drive);
}


int initio_dma_host_off(ide_drive_t *drive)
{
	ide_hwif_t	*hwif=HWIF(drive);
	u8			val;

	// Base5+0B bit 8: DMA transfer disable
	val = hwif->INB(hwif->dma_command);
	hwif->OUTB((val&0x7F), hwif->dma_command);
	return 0;
}


int initio_dma_host_on(ide_drive_t *drive)
{
	if (drive->using_dma)
	{
		ide_hwif_t	*hwif=HWIF(drive);
		u32			addr;	
		u8			val;

		// Base5+0B bit 8: IDMA transfer disable
		addr = initio_selreg(hwif, PRDCTL);
		val = hwif->INB(addr);
		hwif->OUTB((val|0x80), addr);
		return 0;
	}
	return 1;
}


int initio_dma_read(ide_drive_t *drive /*, struct request *rq */)
{
	ide_hwif_t		*hwif=HWIF(drive);
	struct request	*rq=HWGROUP(drive)->rq;
//	ide_task_t 		*args = rq->special;
	unsigned int	count = 0;
	u8 				lba48 = (drive->addressing == 1) ? 1 : 0;
	u32				dma_length = 0;
	task_ioreg_t 	command = WIN_NOP;
	struct scatterlist	*sg = hwif->sg_table;

	if (!(count = ide_build_dmatable(drive, rq, PCI_DMA_FROMDEVICE)))
	{
		printk("   Build PRD table fail!\n");
		/* try PIO instead of DMA */
		return 1;
	}

	/* PRD table */
	hwif->OUTL(hwif->dmatable_dma, hwif->dma_prdtable);

	/* Specify PRD Transfer Length */
	for(count=0;count<hwif->sg_nents;count++)
	{
//		printk("   # PRD Entry %d -> Length=0x%08lX\n", count, (unsigned long)sg_dma_len(sg));
		dma_length += sg_dma_len(sg);
		sg++;
	}
	hwif->OUTL(dma_length, initio_selreg(hwif, PXFRCNT));
//	printk("==> Total: 0x%08lX bytes\n                              -> ", (unsigned long)dma_length);

	/* specify r/w */
	hwif->OUTB(0x88, hwif->dma_command);

	/* clear INTR & ERROR flags */
	hwif->OUTB(0xFF, hwif->dma_status);

	drive->waiting_for_dma = 1;
	if (drive->media != ide_disk)
		return 0;

	// FIX ME to use only ACB ide_task_t args Struct
	command = (lba48) ? WIN_READDMA_EXT : WIN_READDMA;
	if (rq->cmd == IDE_DRIVE_TASKFILE)
	{
		ide_task_t *args = rq->special;
		command = args->tfRegister[IDE_COMMAND_OFFSET];
	}

	/* issue cmd to drive */
	ide_execute_command(drive, command, &initio_dma_intr, 2*WAIT_CMD, initio_timer_expiry);
	
	return HWIF(drive)->ide_dma_count(drive);
}


int initio_dma_write (ide_drive_t *drive /*, struct request *rq */)
{
	ide_hwif_t		*hwif=HWIF(drive);
	struct request	*rq=HWGROUP(drive)->rq;
//	ide_task_t *args	= rq->special;
	unsigned int	count=0;
	u8 				lba48=(drive->addressing == 1) ? 1 : 0;
	u32				dma_length = 0;
	task_ioreg_t	command=WIN_NOP;
	struct scatterlist	*sg = hwif->sg_table;

	printk("   initio_dma_write:\n");

	if (!(count = ide_build_dmatable(drive, rq, PCI_DMA_TODEVICE)))
	{
		printk("      ERROR: ide_build_dmatable fail\n");
		/* try PIO instead of DMA */
		return 1;
	}

	/* PRD table */
	hwif->OUTL(hwif->dmatable_dma, hwif->dma_prdtable);

	/* Specify PRD Transfer Length */
	for(count=0;count<hwif->sg_nents;count++)
	{
		dma_length += sg_dma_len(sg);
		sg++;
	}
	hwif->OUTL(dma_length, initio_selreg(hwif, PXFRCNT));

	printk("      Transfer Length: 0x%08lX bytes\n", (unsigned long)dma_length);

	/* specify r/w */
	hwif->OUTB(0x80, hwif->dma_command);

	/* clear INTR & ERROR flags */
	hwif->OUTB(0xFF, hwif->dma_status);

	drive->waiting_for_dma = 1;
	if (drive->media != ide_disk)
	{
		printk("      ERROR: Non-ide_disk drive\n");
		return 0;
	}
	/*
	 * FIX ME to use only ACB ide_task_t args Struct
	 */
	command = (lba48) ? WIN_WRITEDMA_EXT : WIN_WRITEDMA;
	if (rq->cmd == IDE_DRIVE_TASKFILE)
	{
		ide_task_t *args = rq->special;
		command = args->tfRegister[IDE_COMMAND_OFFSET];
	}
	/* issue cmd to drive */
	printk("      ide_execute_command: 0x%02X\n", command);
	ide_execute_command(drive, command, &initio_dma_intr, 2*WAIT_CMD, initio_timer_expiry);
	return HWIF(drive)->ide_dma_count(drive);
}


/* returns 1 on error, 0 otherwise */
int initio_dma_end(ide_drive_t *drive)
{
	ide_hwif_t	*hwif=HWIF(drive);
	u8			dma_stat=0, dma_cmd=0;

	drive->waiting_for_dma = 0;
	/* get dma_command mode */
	dma_cmd = hwif->INB(hwif->dma_command);

	/* get DMA status */
	dma_stat = hwif->INB(hwif->dma_status);

	/* stop DMA */
	hwif->OUTB(0, hwif->dma_command);
	
	/* clear the INTR & ERROR bits */
	hwif->INB(IDE_STATUS_REG);
	hwif->OUTB(dma_stat, hwif->dma_status);

	/* purge DMA mappings */
	ide_destroy_dmatable(drive);

	/* verify good DMA status */
	dma_stat = hwif->INB(hwif->dma_vendor3);

	return (dma_stat & 0x03)? dma_stat : 0;
}

/*===================================================================
 *	initio_mmio_ide_dma_test_irq - check we caused an IRQ
 *	@drive: drive we are testing
 *
 *	Check if we caused an IDE DMA interrupt. We may also have caused
 *	SATA status interrupts, if so we clean them up and continue.
 *===================================================================*/
static int initio_mmio_ide_dma_test_irq(ide_drive_t *drive)
{
	ide_hwif_t		*hwif=HWIF(drive);
	u8				ata_int=hwif->INB(hwif->dma_status);

	printk("\n%s: initio_mmio_ide_dma_test_irq\n", drive->name);
	printk("     INT Status = 0x%02X\n", ata_int);


	if (SATA_ERROR_REG)
	{
		u32	sata_error=hwif->INL(SATA_ERROR_REG);
		u8	watchdog=0;

//		printk("     SATA Error = 0x%08lX\n", (unsigned long)sata_error);

		if (sata_error & 0x00680000)
		{
			hwif->OUTL(sata_error, SATA_ERROR_REG);
			watchdog = 1;
		}
		if (!(ata_int) && !watchdog)
		{
//			printk("   ### Not our interrupt with SATA_ERROR_REG\n");
			return 0;
		}
	}

	/* return 1 if INTR asserted */
	if (ata_int & 0x94)
	{
//		printk("===> Yes, this is our interrupt!\n");
		return 1;
	}

//	printk("   ### Not our interrupt without SATA_ERROR_REG\n");
	return 0;
}


static int initio_timer_expiry(ide_drive_t *drive)
{
	ide_hwif_t	*hwif=HWIF(drive);
	//u8 		dma_stat=hwif->INB(hwif->dma_status);
	u8 			idma_stat=hwif->INB(hwif->dma_vendor1);

	printk(KERN_WARNING "%s: initio_timer_expiry: dma status == 0x%02x\n", drive->name, idma_stat);

//	if ((dma_stat & 0x18) == 0x18)	/* BUSY Stupid Early Timer !! */
//		return WAIT_CMD;

	/*
	 * Clear the expiry handler in case we decide to wait more,
	 * next time timer expires it is an error 
	 */
	HWGROUP(drive)->expiry = NULL;

	if (idma_stat & 0x03)	/* ERROR */
		return -1;

	if (idma_stat & 0x60)	/* DMAing */
		return WAIT_CMD;

	if (idma_stat & 0x80)	/* Got an Interrupt */
		return WAIT_CMD;

	return 0;	/* Unknown status -- reset the bus */
}


/*******************
 * HWIF Functions  *
 *******************/
/*=========================================================
 *	initio_reset - reset a device on an initio controller
 *	@drive: drive to reset
 *
 *	Perform a controller level reset fo the device. For
 *	SATA we must also check the PHY.
 *=========================================================*/
static void initio_reset(ide_drive_t *drive)
{
	ide_hwif_t	*hwif = HWIF(drive);
	u32			addr;
	u16			val;

	printk("\ninic1622: initio_reset\n   %s:\n", drive->name);

	printk("        1. Disable PCI channel interrupt & Reset ADMA\n");
	addr = initio_selreg(hwif, IDMCTL);
	val = hwif->INW(addr) | aIEN;
	hwif->OUTB((val|aRSTADM), addr);
	udelay(50);
	hwif->OUTB(val, addr);

	printk("        2. Clear interrupt\n");
	hwif->INB(initio_selreg(hwif, 0x07));
	hwif->OUTB(0xFF, initio_selreg(hwif, INTSTAT));

	printk("        3. Enable PCI channel interrupt\n");
	addr = initio_selreg(hwif, IDMCTL);
	val = hwif->INW(addr) & ~(aIEN);
	hwif->OUTB(val, addr);
	
	if (SATA_STATUS_REG)
	{
		u32 sata_stat = hwif->INL(SATA_STATUS_REG);
		printk("   Reset PHY, status=0x%08X\n", sata_stat);
		if (!(sata_stat))
		{
			printk("   Reset PHY Dead\n");
			drive->failures++;
		}
	}

	printk("        - Reset Complete:\n");
	printk("           R1x7        = 0x%02X\n", hwif->INB(IDE_STATUS_REG));
	printk("           R3x6        = 0x%02X\n", hwif->INB(IDE_CONTROL_REG));
	printk("           INT Status  = 0x%02X\n", hwif->INB(initio_selreg(hwif, 0x08)));
	printk("           IDMA CTRL   = 0x%04X\n", hwif->INW(initio_selreg(hwif, 0x14)));
	printk("           IDMA Status = 0x%02X\n", hwif->INB(initio_selreg(hwif, 0x16)));
	printk("           SATA Error  = 0x%08lX\n", (unsigned long)hwif->INL(initio_selreg(hwif, 0x24)));
}


/*====================================================================
 *	initio_tune_chipset	-	set controller timings
 *	@drive: Drive to set up
 *	@xferspeed: speed we want to achieve
 *
 *	Tune the SII chipset for the desired mode. If we can't achieve
 *	the desired mode then tune for a lower one, but ultimately
 *	make the thing work.
 *====================================================================*/
static int initio_tune_chipset (ide_drive_t *drive, byte xferspeed)
{
	ide_hwif_t	*hwif=HWIF(drive);
	u8 			speed, mode=0;
	int			ret;
	unsigned int	val;
	unsigned long	addr;

	printk("hwif->speedproc==initio_tune_chipset:\n");

	if(strstr(drive->id->model, "Maxtor"))
		mode = 3;
	else
		mode = 4;
	speed = ide_rate_filter(mode, xferspeed);

	ret = ide_config_drive_speed(drive, speed);

	// Clear Interrupt
	hwif->INB(initio_selreg(hwif, 0x07));
	hwif->OUTB(0xFF, hwif->dma_status);

	// Negate nIEN in ATL Status Reg
//	hwif->OUTB(0xA0, initio_selreg(hwif, 0x06));
//	hwif->OUTB(0x08, initio_selreg(hwif, ALTSTAT));

	// Enable Channel Interrupt
	addr = initio_selreg(hwif, IDMCTL);
	val = hwif->INW(addr);
	hwif->OUTW(val&0xFEFF, addr);

	printk("               # Enable Channel Interrupt\n");
	printk("                  - R1x7         = 0x%02X\n", hwif->INB(IDE_STATUS_REG));
	printk("                  - PRD Control  = 0x%02X\n", hwif->INB(initio_selreg(hwif, PRDCTL)));
	printk("                  - INT Status   = 0x%02X\n", hwif->INB(initio_selreg(hwif, INTSTAT)));
	printk("                  - IDMA Control = 0x%04X\n", hwif->INW(initio_selreg(hwif, IDMCTL)));
	printk("                  - IDMA Status  = 0x%02X\n", hwif->INB(initio_selreg(hwif, IDMSTAT)));
	printk("                  - SATA ERROR   = 0x%08lX\n", (unsigned long)hwif->INL(SATA_ERROR_REG));

	drive->addressing =  0;
	return ret;
}


/*=========================================================
 *	initio_tuneproc - tune a drive
 *	@drive: drive to tune
 *	@mode_wanted: the target operating mode
 *
 *	Load the timing settings for this device mode into the
 *	controller. If we are in PIO mode 3 or 4 turn on IORDY
 *	monitoring (bit 9). The TF timing is bits 31:16
 *=========================================================*/
static void initio_tuneproc(ide_drive_t *drive, byte mode_wanted)
{
	printk("inic1622: initio_tune_proc\n   %s: Do NOTHING!\n", drive->name);
}


/*===================================================================
 *	initio_reset_poll - wait for sata reset
 *	@drive: drive we are resetting
 *
 *	Poll the SATA phy and see whether it has come back from the dead
 *	yet.
 *===================================================================*/
static int initio_reset_poll(ide_drive_t *drive)
{
	if (SATA_STATUS_REG)
	{
		ide_hwif_t	*hwif = HWIF(drive);

		if ((hwif->INL(SATA_STATUS_REG) & 0x03) != 0x03)
		{
			printk("        Reset PHY dead, status=0x%08x\n", hwif->INL(SATA_STATUS_REG));
			HWGROUP(drive)->poll_timeout = 0;
			return ide_started;
		}
		return 0;
	}
	else
	{
		return 0;
	}
}


/*================================================================
 *	initio_pre_reset - reset hook
 *	@drive: IDE device being reset
 *
 *	For the SATA devices we need to handle recalibration/geometry
 *	differently
 *================================================================*/
static void initio_pre_reset(ide_drive_t *drive)
{
	if (drive->media != ide_disk)
		return;

	drive->special.b.set_geometry = 0;
	drive->special.b.recalibrate = 0;
}


/*=============================================================
 *	initio_busproc - bus isolation ioctl
 *	@drive: drive to isolate/restore
 *	@state: bus state to set
 *
 *	Used by the inic1622 to handle bus isolation. As this is a 
 *	SATA controller the work required is quite limited, we 
 *	just have to clean up the statistics
 *=============================================================*/
static int initio_busproc(ide_drive_t * drive, int state)
{
	ide_hwif_t	*hwif=HWIF(drive);

	switch (state)
	{
		case BUSSTATE_ON:
			hwif->drives[0].failures = 0;
			hwif->drives[1].failures = 0;
			break;
		case BUSSTATE_OFF:
			hwif->drives[0].failures = hwif->drives[0].max_failures + 1;
			hwif->drives[1].failures = hwif->drives[1].max_failures + 1;
			break;
		case BUSSTATE_TRISTATE:
			hwif->drives[0].failures = hwif->drives[0].max_failures + 1;
			hwif->drives[1].failures = hwif->drives[1].max_failures + 1;
			break;
		default:
			return -EINVAL;
	}
	hwif->bus_state = state;
	return 0;
}


/*=================================================================
 *	initio_dma_intr - IDE DMA interrupt handler
 *	@drive: the drive the interrupt is for
 *
 *	Handle an interrupt completing a read/write DMA transfer on an 
 *	IDE device
 *=================================================================*/
ide_startstop_t initio_dma_intr(ide_drive_t *drive)
{
	ide_hwif_t	*hwif=HWIF(drive);
	u8			stat=0, dma_stat=0;
	int			i;

	for(i=0;i<500;i++)
	{
		if ((hwif->INB(hwif->dma_status)&0x10) || (hwif->INB(IDE_STATUS_REG)==0x50))
			break;
		udelay(10);
	}

	/***** Jason Debug *****/
	printk("%s: initio_dma_intr (polling %d)\n", drive->name, i);
	printk("   R1x7       = 0x%02X\n   INT Status = 0x%02X\n", 
			hwif->INB(IDE_STATUS_REG), hwif->INB(hwif->dma_status));

	dma_stat = hwif->ide_dma_end(drive);
	stat = hwif->INB(IDE_STATUS_REG);	/* get drive status */

	if (stat & 0x01)
	{
		printk("   R3x6        = 0x%02X\n", hwif->INB(IDE_CONTROL_REG));
		printk("   Error Reg   = 0x%02X\n", hwif->INB(IDE_ERROR_REG));
	}

	if (dma_stat)
		printk("   IDMA Status = 0x%02X\n", hwif->INB(hwif->dma_vendor3));
	printk("========================================\n");

	if (OK_STAT(stat, DRIVE_READY,drive->bad_wstat|DRQ_STAT))
	{
		if (!dma_stat)
		{
			struct request *rq = HWGROUP(drive)->rq;
	//		rq = HWGROUP(drive)->rq;
			for (i = rq->nr_sectors; i > 0;)
			{
				i -= rq->current_nr_sectors;
				DRIVER(drive)->end_request(drive, 1);
			}
			return ide_stopped;
		}
	}
	return DRIVER(drive)->error(drive, "dma_intr", stat);
}


void inic1622_intrproc(int irq, void *dev_id, struct pt_regs *regs)
{
	ide_hwif_t	*hwif = (ide_hwif_t	*)dev_id;
	u8			intr = hwif->INB(hwif->dma_status);

	if (intr & 0x94)	// Command Complete
	{
	}
	else if (intr & 0x03)	// SATA On/Off
	{
	}
	else
		return;
}


/**********************
 * Chipset Functions  *
 **********************/

/*==================================================================
 *	init_chipset_initio - set up an SI device
 *	@dev: PCI device
 *	@name: device name
 *
 *	Perform the initial PCI set up for this device. Attempt to switch
 *	to 133MHz clocking if the system isn't already set up to do it.
 ===================================================================*/
static unsigned int __init init_chipset_initio(struct pci_dev *dev, const char *name)
{
//	u8 tmpbyte	= 0;

	printk("\n%s: init_chipset_initio\n", name);

	// Write CacheLineSize

	// Default memory mapped I/O registers
	if (pci_resource_start(dev, 5))
	{
		if (setup_mmio_initio(dev, name))
		{
			return 0;
		}
	}

	// Set channel transfer mode & Clock

#if defined(DISPLAY_INITIO_TIMINGS) && defined(CONFIG_PROC_FS)
	initio_devs[n_initio_devs++] = dev;

	if (!initio_proc)
	{
		initio_proc = 1;
		ide_pci_register_host_proc(&initio_procs[0]);
	}
#endif /* DISPLAY_INITIO_TIMINGS && CONFIG_PROC_FS */
	return 0;
}


/*==================================================================
 *	init_iops_initio - set up iops
 *	@hwif: interface to set up
 *
 *	Do the basic setup for the INITIO hardware interface
 *	and then do the MMIO setup if we can. This is the first
 *	look in we get for setting up the hwif so that we
 *	can get the iops right before using them.
 ===================================================================*/
static void __init init_iops_initio(ide_hwif_t *hwif)
{
	struct pci_dev	*dev = hwif->pci_dev;
	
	printk("\n===== INIC1622 =====\n");
	printk("%s:\n   1.init_iops_initio\n", hwif->name);

	hwif->hwif_data = 0;	// No extra hwif data
	hwif->rqsize = 128;		// Max. 128KB/Request

	if (pci_get_drvdata(dev) == NULL)
		return;

	printk("        init_mmio_iops_initio:\n");
	printk("           - default HWIF IOPS\n");
	printk("           - set I/O ports & SATA port address\n");
	
	init_mmio_iops_initio(hwif);
}


/*==================================================================
 *	init_dma_initio - set up IDE DMA
 *	@hwif: interface
 *	@dmabase: DMA base address to use
 *	
 *	For the SI chips this requires no special set up so we can just
 *	let the IDE DMA core do the usual work.
 ===================================================================*/
 static void __init init_dma_initio(ide_hwif_t *hwif, unsigned long dmabase)
{
	u8	dma0_stat, dma1_stat;

	printk("   2.init_dma_initio\n");

	/*****************
	 * ide_dma_iobase
	 *****************/
	if (hwif->mmio)
	{
		printk("        MMIO-DMA at 0x%08lx-0x%08lx", dmabase, dmabase + 0x15 - 1);
		if (check_mem_region(dmabase, 0x15))
		{
			printk("   -- Error, MMIO ports already in use.\n");
			return;
		}
		request_mem_region(dmabase, 0x15, hwif->name);
		hwif->dma_base = dmabase;
		if ((hwif->cds->extra) && (hwif->channel == 0))
		{
			printk("????? init_dma_initio: Why here?\n");
//			request_region(dmabase+16, hwif->cds->extra, hwif->cds->name);
//			hwif->dma_extra = hwif->cds->extra;
		}

		/* There is an issue to watch here. The master might not be 
		   registered because the BIOS disabled it. Eventually this should
		   be fixed by always registering the mate */   
		if(hwif->mate == NULL)
			hwif->dma_master = dmabase;
		else
			hwif->dma_master = (hwif->channel) ? hwif->mate->dma_base : dmabase;
	}
	else
	{
		printk("        BM-DMA at 0x%04lx-0x%04lx", dmabase, dmabase + 0x08 - 1);
		if (!request_region(dmabase, 0x08, hwif->name))
		{
			printk(" -- Error, ports in use.\n");
			return;
		}
		hwif->dma_base = dmabase;

		/* There is an issue to watch here. The master might not be 
		   registered because the BIOS disabled it. Eventually this should
		   be fixed by always registering the mate */
		if(hwif->mate == NULL)
			hwif->dma_master = dmabase;
		else
			hwif->dma_master = (hwif->channel) ? hwif->mate->dma_base : dmabase;
	}

	/***************************
	 * ide_allocate_dma_engine
	 ***************************/
	hwif->dmatable_cpu = pci_alloc_consistent(hwif->pci_dev,
											PRD_ENTRIES * PRD_BYTES,
											&hwif->dmatable_dma);
	hwif->sg_table = kmalloc(sizeof(struct scatterlist) * PRD_ENTRIES, GFP_KERNEL);

	if (!(hwif->dmatable_cpu) || !(hwif->sg_table))
	{
		printk("   %s: -- Error, unable to allocate%s%s table(s).\n",
				(hwif->dmatable_cpu == NULL) ? " CPU" : "",
				(hwif->sg_table == NULL) ?  " SG DMA" : " DMA",
				hwif->cds->name);
		/*************************
		 * ide_release_dma_engine
		 *************************/
		if (hwif->dmatable_cpu)
		{
			pci_free_consistent(hwif->pci_dev,
								PRD_ENTRIES * PRD_BYTES,
								hwif->dmatable_cpu,
								hwif->dmatable_dma);
			hwif->dmatable_cpu = NULL;
		}
		if (hwif->sg_table)
		{
			kfree(hwif->sg_table);
			hwif->sg_table = NULL;
		}
		/******************
		 * ide_release_dma
		 ******************/
		if (hwif->mmio)
		{
			release_mem_region(hwif->dma_base, 0x16);
		}
		else
		{
			release_region(hwif->dma_base, 0x16);
		}
		return;
	}

	if (!(hwif->dma_command))
		hwif->dma_command	= hwif->dma_base;
	if (!(hwif->dma_status))
		hwif->dma_status	= hwif->dma_base - 2 ;

	if (!(hwif->dma_vendor1))
		hwif->dma_vendor1	= (hwif->dma_base + 9);		//IDMA Control
	if (!(hwif->dma_vendor3))
		hwif->dma_vendor3	= (hwif->dma_base + 11);	//IDMA Status
	if (!(hwif->dma_prdtable))
		hwif->dma_prdtable	= (hwif->dma_base + 1);

	// Hook ISR - 2/11/2004
	hwif->intrproc = &inic1622_intrproc;

	// DMA Function
	if (!hwif->ide_dma_off)
		hwif->ide_dma_off = &__ide_dma_off;
	if (!hwif->ide_dma_off_quietly)
		hwif->ide_dma_off_quietly = &__ide_dma_off_quietly;
	if (!hwif->ide_dma_host_off)
		hwif->ide_dma_host_off = &initio_dma_host_off;
	if (!hwif->ide_dma_on)
		hwif->ide_dma_on = &__ide_dma_on;
	if (!hwif->ide_dma_host_on)
		hwif->ide_dma_host_on = &initio_dma_host_on;
	if (!hwif->ide_dma_count)
		hwif->ide_dma_count = &__ide_dma_count;
	if (!hwif->ide_dma_begin)
		hwif->ide_dma_begin = &__ide_dma_begin;
	if (!hwif->ide_dma_end)
		hwif->ide_dma_end = &initio_dma_end;

	if (!hwif->ide_dma_read)
		hwif->ide_dma_read = &initio_dma_read;
	if (!hwif->ide_dma_write)
		hwif->ide_dma_write = &initio_dma_write;

	if (!hwif->ide_dma_bad_drive)
		hwif->ide_dma_bad_drive = &__ide_dma_bad_drive;
	if (!hwif->ide_dma_good_drive)
		hwif->ide_dma_good_drive = &__ide_dma_good_drive;
	if (!hwif->ide_dma_timeout)
		hwif->ide_dma_timeout = &__ide_dma_timeout;
	if (!hwif->ide_dma_retune)
		hwif->ide_dma_retune = &__ide_dma_retune;
	if (!hwif->ide_dma_lostirq)
		hwif->ide_dma_lostirq = &__ide_dma_lostirq;

	dma0_stat = hwif->INB(hwif->dma_base);
	dma1_stat = hwif->INB(hwif->dma_base+0x40);
	printk(", BIOS settings: %s:%s, %s:%s",
	       hwif->drives[0].name, (dma0_stat & 0x80) ? "DMA" : "pio",
	       hwif->drives[1].name, (dma1_stat & 0x80) ? "DMA" : "pio");
	printk("\n");

	printk("        hwif->hwif_data  = 0x%08lX\n", (unsigned long)hwif->hwif_data);
	printk("        hwif->dma_base   = 0x%08lX\n", (unsigned long)hwif->dma_base);
	printk("        hwif->dma_status = 0x%08lX\n", (unsigned long)hwif->dma_status);
	printk("        hwif->dma_vendor1= 0x%08lX\n", (unsigned long)hwif->dma_vendor1);
	printk("        hwif->dma_vendor3= 0x%08lX\n", (unsigned long)hwif->dma_vendor3);
	printk("        SATA Status      = 0x%08lX\n", initio_selreg(hwif, SSTATUS));
	printk("        SATA ERROR       = 0x%08lX\n", initio_selreg(hwif, SERROR));
	
	if (!(hwif->dma_master))
		BUG();
}


/*===================================================================
 *	init_hwif_initio - set up hwif structs
 *	@hwif: interface to set up
 *
 *	We do the basic set up of the interface structure. The INITIO
 *	requires several custom handlers so we override the default
 *	ide DMA handlers appropriately
 ====================================================================*/
 static void __init init_hwif_initio(ide_hwif_t *hwif)
{
	unsigned long	addr, sata_status;
	unsigned int	val;

	printk("\n   3.init_hwif_initio\n");

	hwif->autodma = 0;
	
	hwif->resetproc = &initio_reset;
	hwif->speedproc = &initio_tune_chipset;
	hwif->tuneproc	= &initio_tuneproc;
	hwif->reset_poll = &initio_reset_poll;
	hwif->pre_reset = &initio_pre_reset;
	hwif->busproc   = &initio_busproc;

	if (!hwif->dma_base) {
		hwif->drives[0].autotune = 1;
		hwif->drives[1].autotune = 1;
		return;
	}

	hwif->ultra_mask = 0x7f;
	hwif->mwdma_mask = 0x07;
	hwif->swdma_mask = 0x07;
	hwif->atapi_dma = 1;
	hwif->ide_dma_check = &initio_config_drive_for_dma;

	if (!(hwif->udma_four))
		hwif->udma_four = 1;

	hwif->ide_dma_test_irq = &initio_mmio_ide_dma_test_irq;
	hwif->ide_dma_verbose = &__ide_dma_verbose;
	
	/*
	 *	The BIOS often doesn't set up DMA on this controller
	 *	so we always do it.
	 */

	hwif->autodma = 1;
	hwif->drives[0].autodma = hwif->autodma;
	hwif->drives[1].autodma = hwif->autodma;
	
	//=== Clear ATA Interrupt Status Register
//	printk("      - Clear Channel Interrupt\n");
//	hwif->OUTB(0xFF, initio_selreg(hwif, INTSTAT));

	//=== Disable Channel Interrupt & Reset ADMA
	printk("      - Disable Channel Interrupt & Reset ADMA\n");
	addr = initio_selreg(hwif, IDMCTL);
	val = hwif->INW(addr) | aIEN;
	hwif->OUTW((val|aRSTADM), addr);
	udelay(50);
	hwif->OUTW(val, addr);

	//=== Disable Device INT
//	printk("      - Disable Device INT\n");
//	hwif->OUTB(0x0A, initio_selreg(hwif, ALTSTAT));

	//=== Clear ATA Interrupt Status Register
	printk("      - Clear Interrupt\n");
	hwif->INB(initio_selreg(hwif, 0x07));
	hwif->OUTB(0xFF, initio_selreg(hwif, INTSTAT));

	//=== Set ATA Mask Interrupt Status Register
	printk("      - Set INT Mask\n");
	hwif->OUTB(0x20, initio_selreg(hwif, MINTSTAT));

	//=== Disable IDMA Control aGO bit
	printk("      - Clear aGO bit in IDMA CTRL\n");
	addr = initio_selreg(hwif, IDMCTL);
	val = hwif->INW(addr) & (~aGO);
	hwif->OUTW(val, addr);

	//=== Detect Device
	sata_status = (unsigned long)hwif->INL(initio_selreg(hwif, SSTATUS));
	printk("      - SATA Status  = 0x%08lX\n", sata_status);
	if ((sata_status&0x03) == 0x03)
		hwif->drives[0].present = 1;
	else
		hwif->drives[0].present = 0;
	hwif->drives[1].noprobe = 1;

	printk("      - R1x7         = 0x%02X\n", hwif->INB(initio_selreg(hwif, 0x07)));
	printk("      - INT Status   = 0x%02X\n", hwif->INB(hwif->dma_status));
	printk("      - PRD Control  = 0x%02X\n", hwif->INB(hwif->dma_command));
	printk("      - IDMA Control = 0x%04X\n", hwif->INW(hwif->dma_vendor1));
	printk("      - IDMA Status  = 0x%04X\n", hwif->INW(hwif->dma_vendor3));
	printk("      - SATA Error   = 0x%08lX\n", (unsigned long)hwif->INL(addr));

	// Allocate an interrupt line - 2/11/2004
	request_irq(hwif->irq, &inic1622_intrproc, SA_SHIRQ, hwif->name, hwif);
	
}


extern void ide_setup_pci_device(struct pci_dev *, ide_pci_device_t *);
/*==================================================================
 *	initio_init_one - pci layer discovery entry
 *	@dev: PCI device
 *	@id: ident table entry
 *
 *	Called by the PCI code when it finds an Inic1622 controller.
 *	We then use the IDE PCI generic helper to do most of the work.
 ===================================================================*/
static int __devinit initio_init_one(struct pci_dev *dev, const struct pci_device_id *id)
{
	ide_pci_device_t *d = &initio_chipsets[id->driver_data];
	if (dev->device != d->device)
		BUG();

	printk("\n%s: initio_init_one\n", dev->name);

	ide_setup_pci_device(dev, d);
	MOD_INC_USE_COUNT;
	return 0;
}




static struct pci_device_id initio_pci_tbl[] __devinitdata = {
	{ PCI_VENDOR_ID_INIT, PCI_DEVICE_ID_INIT_1622,  PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0},
	{ 0, },
};

static struct pci_driver driver = {
	.name		= "Initio SATA",
	.id_table	= initio_pci_tbl,
	.probe		= initio_init_one,
};

static int initio_ide_init(void)
{
	printk("inic1622: module_init\n");
	return ide_pci_register_driver(&driver);
}

static void initio_ide_exit(void)
{
	printk("inic1622: module_exit\n");
	ide_pci_unregister_driver(&driver);
}

module_init(initio_ide_init);
module_exit(initio_ide_exit);

MODULE_AUTHOR("Andre Hedrick, Alan Cox, Jason Lai");
MODULE_DESCRIPTION("PCI driver module for Initio SATA");
MODULE_LICENSE("GPL");
